import tempfile
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field, confloat, root_validator
from pydantic.dataclasses import dataclass
from pydantic.error_wrappers import ValidationError

from osp.core.cuds import Cuds
from osp.core.namespaces import emmo
from osp.core.session import CoreSession
from osp.core.utils import export_cuds
from osp.models.utils.general import _get_example, get_download, get_upload
from osp.wrappers.simcatalyticfoam.utils import (
    _make_boundary_field,
    _make_internal_field,
    _make_value,
)

from .utils import _import_from_pkl, _read_patches, make_arcp

if TYPE_CHECKING:
    from typing import List, Union

    from osp.core.cuds import Cuds
    from osp.core.ontology import OntologyClass


class Boundary(str, Enum):
    """Model with a chosen set of boundary types."""

    fixedValue = "FixedValue"
    fixedGradient = "FixedGradient"
    catalyticWall = "CatalyticWall"
    empty = "Empty"
    zeroGradient = "ZeroGradient"
    processor = "Processor"
    patch = "Patch"
    SymmetryPlane = "SymmetryPlane"
    wedge = "Wedge"
    wall = "Wall"


class TurbulenceModel(str, Enum):
    """Turblence model of the CFD-calculation."""

    kOmegaSST = "KOmegaSST"
    kEpsilon = "KEpsilon"
    laminar = "LaminarModel"
    smagorinsky = "SmagorinskyTurbulenceModel"
    kinetic_energy_equation = "KineticEnergyEquation"


class DiffusivityModel(str, Enum):
    ficks_diffusion = "FicksDiffusion"
    maxwell_diffusion = "MaxwellStefanDiffusion"


class ChemicalComposition(str):
    """Chemical composition of the species."""

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(standard_species=_import_from_pkl())


class Patch(str):
    """Available patches in the regarded use case."""

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(standard_patches=_read_patches())


class BoundaryCondition(BaseModel):
    """Patch and boundary type definition."""

    boundary_type: Boundary = Field(
        Boundary.zeroGradient.value,
        description="Boundary types available in OpenFOAM (most popular ones)",
    )
    patch: Patch = Field(
        ..., description="Patch in the mesh available from the use case."
    )
    value: Optional[Union[float, int, List[confloat(allow_inf_nan=False)]]] = Field(
        None,
        description="""Fixed value (Dirichlet-condition),
        fixed gradient (von-Neumann-condition) or field value at reacting catalyst wall.
        Field will be ignored for other conditions.""",
    )

    @root_validator
    def validate_all(cls, values):
        if values.get("boundary_type") in [
            Boundary.fixedValue,
            Boundary.fixedGradient,
            Boundary.catalyticWall,
        ] and not values.get("value"):
            raise ValidationError(
                """If `boundary` is set to `fixedValue`, `fixedGradient` or `catalyticWall`,
                `value`-field must be set as well."""
            )
        return values

    class Config:
        use_enum_values = True


class BaseQuantity(BaseModel):
    """BaseQuantity for vector and scalar fields."""

    boundaries: List[BoundaryCondition] = Field(
        ...,
        description="""Mapping between available patches and boundary condition.
        NOTE: All patches in the mesh must be defined!""",
    )

    class Config:
        use_enum_values = True


class SolverOptions(BaseModel):
    """Solver options model for diffusivity and turbulence properties."""

    diffusivity_model: DiffusivityModel = Field(
        DiffusivityModel.ficks_diffusion.value,
        description="Diffusivity of the mixture: Fick's multi component diffusion or Maxwell Stefan diffusion.",
    )
    turbulence_model: TurbulenceModel = Field(
        TurbulenceModel.laminar.value,
        description="Turbulence model of the mixture: laminar, RAS or LES.",
    )
    use_energy_equation: bool = Field(
        False, description="Whether the energy equation should be regarded or not."
    )

    class Config:
        use_enum_values = True


class Velocity(BaseQuantity):
    """Velocity model of the mixture."""

    value: List[confloat(allow_inf_nan=False)] = Field(
        ..., description="Velocity vector in m/s.", min_items=3, max_items=3
    )


class Temperature(BaseQuantity):
    """Temperature model of the mixture."""

    value: confloat(allow_inf_nan=False) = Field(
        ..., description="Temperature value of the mixture in K."
    )


class Pressure(BaseQuantity):
    """Pressure model of the mixture."""

    value: confloat(allow_inf_nan=False) = Field(
        ..., description="Pressure value of the mixture in Pa."
    )


class ChemicalSpecies(BaseQuantity):
    """A chemical species with defined properties."""

    composition: ChemicalComposition = Field(
        ...,
        description="""Chemical composition of the species.
        Available subset is forwarded from the standard surrogate model.""",
    )

    mass_fraction: confloat(ge=0, le=1, allow_inf_nan=False) = Field(
        ...,
        description="""Mass fraction of the species within the composition.
        Sum of all mass fractions must be 1.""",
    )


@dataclass
class COCatalyticFOAMModel:
    chemical_species: List[ChemicalSpecies] = Field(
        ..., description="List of chemical species involed into the reaction."
    )
    species_from_upload: Optional[UUID] = Field(
        None,
        description="UUID of the PKL in the cache to be used instead of the standard model",
    )
    patches_from_upload: Optional[UUID] = Field(
        None,
        description="If defined, UUID of the use case tar-ball in the cache to be used instead of the standard use case",
    )
    fraction_sum_error: Optional[int] = Field(
        3,
        description="""Number of decimal places which shall be
    used while rounding sum of mass fractions.
    E.g. the test passes when CO=0.4468 and O2=0.5531 with an error 3 digits,
    but fails for 4 digits.""",
    )
    velocity: Velocity = Field(..., description="Velocity model of the mixture.")
    pressure: Pressure = Field(..., description="Pressure model of the mixture.")
    temperature: Temperature = Field(
        ..., description="Temperature model of the mixture."
    )
    catalyst_amount: confloat(ge=0, le=1, allow_inf_nan=False) = Field(
        ..., description="Amount fraction of catalyst in the reactive wall patch."
    )
    solver_options: SolverOptions = Field(
        SolverOptions(),
        description="Solver options including turbulence and diffusion properties.",
    )

    maxtime: confloat(allow_inf_nan=False) = Field(
        1e-7,
        description="Maximum simulation time in seconds until the simulation should run.",
    )
    delta_t: confloat(allow_inf_nan=False) = Field(
        1e-8,
        description="""Inital time step length in seconds.
    Will be adjusted by the solver when `adjustable_timestep` is set to `True`.
    IMPORTANT: For simplicity, the delta_t is going to be the `writeInterval` of the output files!""",
    )
    adjustable_timestep: bool = Field(
        True,
        description="""Whether the time step should be adjusted according to the Courant–Friedrichs–Lewy-condition
                                     in order to keep the stability of the simulation.""",
    )

    @property
    def session(self) -> CoreSession:
        return self._session

    def __post_init_post_parse__(self):
        with CoreSession() as session:
            calc = emmo.ContinuumCalculation()
            mappings = [
                ("0/p", "pressure", emmo.Pressure),
                ("0/T", "temperature", emmo.ThermodynamicTemperature),
                ("0/U", "velocity", emmo.Velocity),
            ]
            for path, attr, oclass in mappings:
                model = getattr(self, attr)
                quantity = self._make_quantity(path, model, oclass)
                calc.add(quantity, rel=emmo.hasCalculationInput)
            for species in self.chemical_species:
                calc.add(self._make_species(species), rel=emmo.hasCalculationInput)
            calc.add(self._make_continuum_model(), rel=emmo.hasCalculationInput)
            calc.add(self._make_catalyist_amount(), rel=emmo.hasCalculationInput)
        file = tempfile.NamedTemporaryFile(suffix=".ttl")
        export_cuds(session, file.name)
        self._uuid = get_upload(file)
        self._session = session

    def _make_time_constraints(self) -> List[Cuds]:
        if self.maxtime < self.delta_t:
            raise ValueError("`maxtime` must be larger equal `delta_t`!")
        quantities = []
        for oclass, attr, key in [
            (emmo.MaximumSimulationTime, "maxtime", "endTime"),
            (emmo.DeltaSimulationTime, "delta_t", "deltaT"),
            (emmo.SimulationWriteOutputInterval, "delta_t", "writeInterval"),
        ]:
            quantity = oclass()
            value = _make_value("system/controlDict", getattr(self, attr), [key])
            quantity.add(value, rel=emmo.hasQuantityValue)
            quantities.append(quantity)
        return quantities

    def _make_species(self, species: ChemicalSpecies) -> Cuds:
        path = f"0/{species.composition}"
        oclass = emmo.MassFraction
        composition = emmo.ChemicalComposition()
        cspecies = emmo.ChemicalSpecies()
        iri = make_arcp(path, query=dict(jsonpath=[["FoamFile", "object"]]))
        symbol = emmo.Symbol(iri=iri, hasSymbolData=species.composition)
        cspecies.add(symbol)
        mass_fraction = _make_internal_field(path, species.mass_fraction, oclass)
        composition.add(mass_fraction, cspecies, rel=emmo.hasSpatialDirectPart)
        for boundary in species.boundaries:
            bc = _make_boundary_field(
                path,
                boundary.patch,
                emmo[boundary.boundary_type],
                value=boundary.value,
                quantity=oclass,
            )
            composition.add(bc, rel=emmo.hasBoundaryField)
        return composition

    def _make_quantity(
        self, path: "str", quantitymodel: "BaseQuantity", oclass: "OntologyClass"
    ) -> "Cuds":
        quantity = _make_internal_field(path, quantitymodel.value, oclass)
        for boundary in quantitymodel.boundaries:
            bc = _make_boundary_field(
                path,
                boundary.patch,
                emmo[boundary.boundary_type],
                value=boundary.value,
                quantity=oclass,
            )
            quantity.add(bc, rel=emmo.hasBoundaryField)
        return quantity

    def _make_continuum_model(self) -> "Cuds":
        model = emmo.ContinuumModel()

        if self.solver_options.use_energy_equation:
            iri = make_arcp(
                "constant/solverOptions",
                query=dict(jsonpath=[["PhysicalModel", "energyEquation"]]),
            )
            energy_eq = emmo.EnergyEquation(iri=iri)
            model.add(energy_eq, rel=emmo.hasSpatialDirectPart)

        iri = make_arcp(
            "constant/solverOptions",
            query=dict(jsonpath=[["PhysicalModel", "diffusivityModel"]]),
        )
        diffusivity_model = emmo[self.solver_options.diffusivity_model](iri=iri)
        model.add(diffusivity_model, rel=emmo.hasSpatialDirectPart)

        model.add(self._make_turbulent_model(), rel=emmo.hasSpatialDirectPart)
        model.add(*self._make_time_constraints(), rel=emmo.hasSpatialDirectPart)

        if self.adjustable_timestep:
            iri = make_arcp(
                "system/controlDict",
                query=dict(jsonpath=[["adjustTimeStep"]]),
            )
            adjust = emmo.AdjustableSimulationTimeStep(iri=iri)
            model.add(adjust, rel=emmo.hasSpatialDirectPart)
        return model

    def _make_turbulent_model(self) -> "List[Cuds]":
        oclass = emmo[self.solver_options.turbulence_model]
        if oclass.is_subclass_of(emmo.ReynoldsAveragedSimulationModel):
            path = ["RAS", "RASModel"]
        elif oclass.is_subclass_of(emmo.LargeEddySimulationModel):
            path = ["LES", "LESModel"]
        elif oclass.is_subclass_of(emmo.LaminarModel):
            path = ["simulationType"]
        else:
            raise TypeError(
                f"""
            Turbulence model `{oclass}` not understood.
            Must be a subclass of `{emmo.ReynoldsAveragedSimulationModel}`,
            `{emmo.LargeEddySimulationModel}` or `{emmo.LaminarModel}`."""
            )
        iri = make_arcp("constant/turbulenceProperties", query=dict(jsonpath=[path]))
        return oclass(iri=iri)

    def _make_catalyist_amount(self) -> "Cuds":
        iri = make_arcp(
            "constant/solverOptions",
            query=dict(jsonpath=[["PhysicalModel", "alfaCatalyst"]]),
        )
        catalyst_amount = emmo.CatalystAmount()
        catalyst_amount_value = emmo.Real(
            iri=iri, hasNumericalData=self.catalyst_amount
        )
        catalyst_amount.add(catalyst_amount_value, rel=emmo.hasQuantityValue)
        return catalyst_amount

    @root_validator
    def validate_all(cls, values):
        if not values.get("chemical_species"):
            raise ValidationError("`chemical_species` is a mandatory field.")
        species_upload = values.get("species_from_upload")
        patches_upload = values.get("patches_from_upload")
        if not species_upload:
            compositions = _import_from_pkl()
        else:
            compositions = cls._get_species_from_upload(species_upload)
        if not patches_upload:
            patches = _read_patches()
        else:
            patches = cls._get_patches_from_upload(patches_upload)
        cls._validate_species(values, compositions)
        quantities = [
            values.get(key) for key in ["temperature", "pressure", "velocity"]
        ]
        species = [species for species in values.get("chemical_species")]
        cls._check_bounds(species + quantities, patches)
        return values

    @classmethod
    def _validate_species(
        self, values: "Dict[Any, Any]", compositions: "List[str]"
    ) -> None:
        sum_species = 0
        for species in values.get("chemical_species"):
            if species.composition not in compositions:
                raise ValueError(
                    f"Invalid composition `{species.composition}`. Please choose from {compositions}"
                )
            sum_species += species.mass_fraction
        error = values.get("fraction_sum_error")
        rounded = round(sum_species, error)
        if rounded != 1.0:
            raise ArithmeticError(
                f"""The sum of `composition` of `chemical_species` must be `1.0`,
            but it was {rounded}, according to an error of `sum_fraction_error`={error}
            (true unrounded sum is: `{sum_species}`).
            Please decrease of your `sum_fraction_error` or
            check the `composition`-values of your `chemical_species`."""
            )

    @classmethod
    def _check_bounds(self, quantities: "List[Any]", patches: "List[str]") -> None:
        for quantity in quantities:
            list_patches = []
            if quantity:
                for condition in quantity.boundaries:
                    list_patches += [condition.patch]
                    if condition.patch not in patches:
                        raise ValueError(
                            f"Invalid patch `{condition.patch}`. Please choose from: `{patches}`"
                        )
            if sorted(patches) != sorted(list_patches):
                raise ValueError(
                    f"""The provided patches for the species `{quantity}` are inconsistent with
                    the list of required patches from the use case: `{patches}`. Your provided
                    patches were: `{list_patches}`. Please complete the boundary conditions with
                    the needed patches.
                    """
                )

    @classmethod
    def _get_species_from_upload(cls, uuid: UUID) -> List[str]:
        file_path = get_download(uuid, as_file=True)
        return _import_from_pkl(file_path)

    @classmethod
    def _get_patches_from_upload(cls, uuid: UUID) -> List[str]:
        file_path = get_download(uuid, as_file=True)
        return _read_patches(case_path=file_path)

    @property
    def cuds(cls):
        return cls._session.load(cls._session.root).first()

    @property
    def uuid(cls):
        return cls._uuid

    class Config:
        """Pydantic config class for CO catalytic FOAM"""

        schema_extra = {"example": _get_example("catalyticFoam", "COModel.json")}



"""Conftest for simphony-catalytic"""
import os
import tempfile
from typing import TextIO, Union
from uuid import uuid4

import pytest

PATH = os.path.dirname(__file__)
TEST_PATH = os.path.join(PATH, "test_files")


def return_file(key: str, as_file=False) -> str:
    """Mocker function for returning test files."""
    tempdir = tempfile.gettempdir()
    path = os.path.join(tempdir, f"{key}.ttl")
    if not os.path.exists(path):
        raise ValueError(f"Key `{key}` does not exist in cache.")
    if as_file:
        return open(path, "rb")
    else:
        return path


def store_file(file: Union[TextIO, str], uuid: str = None) -> str:
    """Mocker for uploading files."""
    tempdir = tempfile.gettempdir()
    if not uuid:
        uuid = str(uuid4())
    path = os.path.join(tempdir, f"{uuid}.ttl")
    with open(path, "wb") as temp:
        temp.write(file.read())
    return uuid


@pytest.fixture
def prepare_env():
    os.environ["REAXPRO_MINIO_USER"] = "foo"
    os.environ["REAXPRO_MINIO_PASSWORD"] = "bar"


@pytest.fixture(autouse=True)
@pytest.mark.usefixtures("mocker")
def mock_download(mocker, prepare_env):
    """Sending request to mock-minio."""
    function = "osp.models.utils.general.get_download"
    patch = mocker.patch(function, new=return_file)
    return patch


@pytest.fixture(autouse=True)
@pytest.mark.usefixtures("mocker")
def mock_upload(mocker, prepare_env):
    """Sending request to mock-minio."""
    function = "osp.models.utils.general.get_upload"
    patch = mocker.patch(function, new=store_file)
    return patch



"""Pytest for the catalyticFOAM with the Pydantic CO-model"""

p_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "p",
    },
    "dimensions": [1, -1, -2, 0, 0, 0, 0],
    "internalField": ["uniform", 100000.0],
    "boundaryField": {
        "inlet": {"type": "zeroGradient"},
        "outlet": {"type": "fixedValue", "value": ["uniform", 100000.0]},
        "inertWall": {"type": "zeroGradient"},
        "reactingWall": {"type": "zeroGradient"},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}

CO_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "CO",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 0.00000945945945945946],
    "boundaryField": {
        "inlet": {"type": "fixedValue", "value": ["uniform", 0.00000945945945945946]},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {
            "type": "zeroGradient",
        },
        "reactingWall": {
            "type": "catalyticWall",
            "value": ["uniform", 0.00000945945945945946],
        },
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}


co2_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "CO2",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 0.567558108108108],
    "boundaryField": {
        "inlet": {"type": "fixedValue", "value": ["uniform", 0.567558108108108]},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {
            "type": "zeroGradient",
        },
        "reactingWall": {
            "type": "catalyticWall",
            "value": ["uniform", 0.567558108108108],
        },
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}


O2_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "O2",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 0.432432432432432],
    "boundaryField": {
        "inlet": {"type": "fixedValue", "value": ["uniform", 0.432432432432432]},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {
            "type": "zeroGradient",
        },
        "reactingWall": {
            "type": "catalyticWall",
            "value": ["uniform", 0.4324324324324328],
        },
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}

U_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volVectorField",
        "location": '"0"',
        "object": "U",
    },
    "boundaryField": {
        "inertWall": {"type": "fixedValue", "value uniform": [0.0, 0.0, 1.0]},
        "inlet": {"type": "fixedValue", "value uniform": [0.0, 0.0, 1.0]},
        "outlet": {"type": "zeroGradient"},
        "reactingWall": {"type": "fixedValue", "value uniform": [0.0, 0.0, 1.0]},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
    "dimensions": [0, 1, -1, 0, 0, 0, 0],
    "internalField": ["uniform", [0.0, 0.0, 1.0]],
}

T_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "T",
    },
    "dimensions": [0, 0, 0, 1, 0, 0, 0],
    "internalField": ["uniform", 900.0],
    "boundaryField": {
        "inlet": {"type": "fixedValue", "value": ["uniform", 900.0]},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {"type": "zeroGradient"},
        "reactingWall": {"type": "catalyticWall", "value": ["uniform", 900.0]},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}


theta_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "thetaDefault",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 0],
    "boundaryField": {
        "inlet": {"type": "zeroGradient"},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {"type": "zeroGradient"},
        "reactingWall": {"type": "fixedValue", "value": ["uniform", 0]},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}

y_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "ascii",
        "class": "volScalarField",
        "location": '"1"',
        "object": "Ydefault",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 0],
    "boundaryField": {
        "reactingWall": {"type": "catalyticWall", "value": ["uniform", 0]},
        "inertWall": {"type": "zeroGradient"},
        "inlet": {"type": "fixedValue", "value": ["uniform", 0]},
        "outlet": {"type": "zeroGradient"},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}

turbulence_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "ascii",
        "class": "dictionary",
        "location": '"constant"',
        "object": "turbulenceProperties",
    },
    "simulationType": "RAS",
    "RAS": {"RASModel": "kEpsilon", "turbulence": "on", "printCoeffs": "on"},
}

solver_expected = {
    "FoamFile": {
        "version": 2.3,
        "format": "binary",
        "class": "dictionary",
        "location": '"constant"',
        "object": "solverOptions",
    },
    "MachineLearning": {
        "MLSurrogate": "on",
        "MachineLearningFile": '"ml_ExtraTrees_forCFD.pkl"',
    },
    "PhysicalModel": {
        "strangAlgorithm": '"TransportReactionMomentum"',
        "homogeneousReactions": "off",
        "heterogeneousReactions": "on",
        "alfaCatalyst": 1.5e-05,
        "Sct": 0.7,
        "Prt": 0.7,
        "catalyticWalls": ["reactingWall"],
        "energyEquation": "on",
        "reactionHeatFromHeterogeneousReactions": "on",
        "constPressureBatchReactor": True,
        "massDiffusionInEnergyEquation": "on",
        "diffusivityModel": "fick-multi-component",
        "includeDpDt": "off",
        "soretEffect": "off",
    },
    "OdeHomogeneous": {
        "odeSolver": '"OpenSMOKE"',
        "relTolerance": 1e-07,
        "absTolerance": 1e-12,
        "maximumOrder": 5,
        "fullPivoting": False,
    },
    "OdeHeterogeneous": {
        "odeSolver": '"OpenSMOKE"',
        "relTolerance": 1e-07,
        "absTolerance": 1e-12,
        "maximumOrder": 5,
        "fullPivoting": False,
    },
    "LewisNumbers": {"O2": 1.0, "H2": 1.1, "H2O": 1.2, "N2": 1.3},
    "#include": '"isatOptions"',
}

control_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "ascii",
        "class": "dictionary",
        "location": '"system"',
        "object": "controlDict",
    },
    "application": "catalyticFoam",
    "startFrom": "startTime",
    "startTime": 0,
    "stopAt": "endTime",
    "endTime": 1e-07,
    "deltaT": 1e-08,
    "writeControl": "adjustableRunTime",
    "writeInterval": 1e-08,
    "purgeWrite": 0,
    "writeFormat": "ascii",
    "writePrecision": 18,
    "writeCompression": "uncompressed",
    "timeFormat": "general",
    "timePrecision": 6,
    "runTimeModifiable": "yes",
    "adjustTimeStep": "yes",
    "maxCo": 0.1,
    "libs": ['"libCatalyticWall.so"'],
}


def test_foam_model():
    import os
    import tempfile

    from osp.core.namespaces import cuba
    from osp.core.utils import import_cuds
    from osp.models.catalytic.co_catalyticfoam import (
        COCatalyticFOAMModel,
        TurbulenceModel,
    )
    from osp.wrappers.simcatalyticfoam.parsers import run_parser
    from osp.wrappers.simcatalyticfoam.simcatalyticfoam import SimCatalyticFoamSession

    from .conftest import return_file
    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config = {
        "chemical_species": [
            {
                "composition": "O2",
                "mass_fraction": 0.432432432432432,
                "boundaries": chem_bounds(0.432432432432432),
            },
            {
                "composition": "CO",
                "mass_fraction": 0.00000945945945945946,
                "boundaries": chem_bounds(0.00000945945945945946),
            },
            {
                "composition": "CO2",
                "mass_fraction": 0.567558108108108,
                "boundaries": chem_bounds(0.567558108108108),
            },
        ],
        "velocity": {
            "value": [0.0, 0.0, 1.0],
            "boundaries": u_bounds([0.0, 0.0, 1.0]),
        },
        "temperature": {"value": 900.0, "boundaries": t_bounds(900.0)},
        "pressure": {"value": 100000.0, "boundaries": p_bounds(100000.0)},
        "catalyst_amount": 1.5e-05,
        "solver_options": {
            "use_energy_equation": True,
            "turbulence_model": TurbulenceModel.kEpsilon,
        },
    }
    model = COCatalyticFOAMModel(**config)
    cuds_file = return_file(model.uuid)

    # check if catalyticFoam is installed
    config = dict(
        commands=[
            dict(
                exec="echo",
                args="hello world",
            )
        ],
        directory=tempfile.TemporaryDirectory().name,
        filename=f"{tempfile.NamedTemporaryFile().name}.tar",
    )

    session = SimCatalyticFoamSession(config=config)
    wrapper = cuba.Wrapper(session=session)
    cuds = import_cuds(cuds_file, session=session)
    wrapper.add(*cuds, rel=cuba.relationship)
    session.run()

    p = run_parser(os.path.join(session.config["directory"], "0", "p"))
    T = run_parser(os.path.join(session.config["directory"], "0", "T"))
    U = run_parser(os.path.join(session.config["directory"], "0", "U"))
    CO = run_parser(os.path.join(session.config["directory"], "0", "CO"))
    co2 = run_parser(os.path.join(session.config["directory"], "0", "CO2"))
    y = run_parser(os.path.join(session.config["directory"], "0", "Ydefault"))
    theta = run_parser(os.path.join(session.config["directory"], "0", "thetaDefault"))
    turbulence = run_parser(
        os.path.join(session.config["directory"], "constant", "turbulenceProperties")
    )
    solver = run_parser(
        os.path.join(session.config["directory"], "constant", "solverOptions")
    )
    control = run_parser(
        os.path.join(session.config["directory"], "system", "controlDict")
    )

    assert p == p_expected

    assert T == T_expected

    assert CO == CO_expected

    assert co2 == co2_expected

    assert theta == theta_expected

    assert y == y_expected

    assert U == U_expected

    assert solver == solver_expected

    assert turbulence == turbulence_expected

    assert control == control_expected

    assert session.exit_code == 0



import logging
from typing import TYPE_CHECKING
from urllib.parse import parse_qs

from osp.core.namespaces import emmo
from osp.core.session import SimWrapperSession
from osp.models.catalytic.utils import check_arcp, wrap_arcp

from .catalyticfoam_engine import CatalyticFoamEngine, settings

if TYPE_CHECKING:
    from typing import Any, List

    from osp.core.cuds import Cuds
    from osp.core.ontology import OntologyClass

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class SimCatalyticFoamSession(SimWrapperSession):
    def __init__(
        self, case: "str" = "laminar_2D_ML", config: "dict" = {}, *args, **kwargs
    ):
        engine = CatalyticFoamEngine(case, config=config)
        super().__init__(engine=engine)

    def __str__(self):
        return "CatalyticFoamSession"

    # OVERRIDE
    def _run(self, root_object) -> None:
        self._engine.run()

    # OVERRIDE
    def _apply_added(self, root_object, buffer) -> None:
        for obj in buffer.values():
            self._wrap(obj)

    # OVERRIDE
    def _wrap(self, cuds: "Cuds") -> None:
        if cuds.is_a(emmo.ChemicalComposition):
            logger.info(f"Found a {emmo.ChemicalComposition}.")
            self._make_new_chemical(cuds)
            self._check_quantity(cuds, emmo.MassFraction)
            self._check_for_boundaries(cuds)
        elif (
            cuds.is_a(emmo.Pressure)
            or cuds.is_a(emmo.ThermodynamicTemperature)
            or cuds.is_a(emmo.Velocity)
        ):
            logger.info(f"Found a {cuds.oclass}.")
            self._check_for_value(cuds)
            self._check_for_boundaries(cuds)
        elif cuds.is_a(emmo.CatalystAmount):
            self._check_for_value(cuds)
        elif (
            cuds.is_a(emmo.EnergyEquation)
            or cuds.is_a(emmo.DiffusivityModel)
            or cuds.is_a(emmo.ProbabilityDistribution)
        ):
            self._wrap_with_label(cuds)
        elif cuds.is_a(emmo.TurbulentModel) or cuds.is_a(emmo.LaminarModel):
            self._check_turbulent_model(cuds)
        elif cuds.is_a(emmo.SimulationTime) and not cuds.is_a(
            emmo.AdjustableSimulationTimeStep
        ):
            self._check_for_value(cuds)
        elif cuds.is_a(emmo.AdjustableSimulationTimeStep):
            self._wrap_arcp(cuds, "yes")

    def _make_new_chemical(self, composition: "Cuds") -> None:
        species = composition.get(oclass=emmo.ChemicalSpecies)
        if species:
            symbol = species.pop().get(oclass=emmo.Symbol)
        if symbol:
            symbol = symbol.pop()
        if check_arcp(symbol):
            logger.info(
                f"Found a {emmo.ChemicalSpecies} for <{composition}> with a {emmo.Symbol}: {symbol.hasSymbolData}."
            )
            self._wrap_arcp(symbol, symbol.hasSymbolData)

    def _check_for_boundaries(self, cuds: "Cuds") -> None:
        boundaries = cuds.get(rel=emmo.hasBoundaryField)
        for obj in boundaries:
            if check_arcp(obj):
                self._wrap_with_label(obj)
            for value in obj.get(rel=emmo.hasSpatialDirectPart):
                self._check_for_value(value)

    def _check_turbulent_model(self, cuds: "Cuds") -> None:
        self._wrap_with_label(cuds)
        if cuds.is_a(emmo.ReynoldsAveragedSimulationModel):
            stype = "RAS"
        elif cuds.is_a(emmo.LargeEddySimulationModel):
            stype = "LES"
        elif cuds.is_a(emmo.LaminarModel):
            stype = None
        else:
            raise TypeError(
                f"""
            Turbulence model `{cuds}` not understood.
            Must be a subclass of `{emmo.ReynoldsAveragedSimulationModel}`,
            `{emmo.LargeEddySimulationModel}` or `{emmo.LaminarModel}`."""
            )
        if stype:
            self.add_to_parser("constant/turbulenceProperties", "simulationType", stype)
            self.add_to_parser(
                "constant/turbulenceProperties", f"{stype}.turbulence", "on"
            )
            self.add_to_parser(
                "constant/turbulenceProperties", f"{stype}.printCoeffs", "on"
            )

    def _wrap_with_label(self, entity: "Cuds") -> None:
        label = self._get_label(entity)
        logger.info(
            f"Found <{entity}> with {label} by annotation {settings.catalyticfoam_label_iri}."
        )
        self._wrap_arcp(entity, label)

    def _get_label(self, obj: "Cuds") -> "str":
        labels = [
            str(o)
            for s, p, o in obj.oclass.get_triples()
            if s == obj.oclass.iri and str(p) == settings.catalyticfoam_label_iri
        ]
        if len(labels) > 1:
            raise ValueError(
                f"More than one OpemFOAM-label for {obj} detected: {labels}. Please check ontology-class."
            )
        elif len(labels) == 0:
            raise ValueError(
                f"No OpemFOAM-label for {obj} detected. Please check ontology-class."
            )
        return labels.pop()

    def _wrap_arcp(self, value: "Cuds", data: "str") -> None:
        result = wrap_arcp(value)
        path = result.path[1:]
        for query in parse_qs(result.query).get("jsonpath"):
            self.add_to_parser(path, query, data)

    def _check_quantity(self, cuds: "Cuds", oclass: "OntologyClass") -> None:
        cuds_list = cuds.get(oclass=oclass)
        logger.info(f"Check <{cuds}> for a related <{oclass}>.")
        if cuds_list:
            self._check_for_value(cuds_list.pop())

    def _check_for_value(self, quantity: "Cuds") -> None:
        value = quantity.get(rel=emmo.hasQuantityValue)
        if value:
            value = value.pop()
            if value.is_a(emmo.Number):
                self._check_number(quantity, value)
            elif value.is_a(emmo.Shape3Vector):
                self._check_vector(quantity, value)

    def _check_number(self, quantity: "Cuds", value: "Cuds") -> None:
        if check_arcp(value):
            logger.info(
                f"Found quantity <{quantity}> with <{value}> valued {value.hasNumericalData}."
            )
            self._wrap_arcp(value, value.hasNumericalData)

    def _check_vector(self, quantity: "Cuds", vector: "Cuds") -> None:
        first = vector.get(rel=emmo.hasSpatialFirst)
        if first:
            first = first.pop()
            self._check_number(quantity, first)
            sec = first.get(rel=emmo.hasSpatialNext)
            if sec:
                sec = sec.pop()
                self._check_number(quantity, sec)
                third = sec.get(rel=emmo.hasSpatialNext)
                if third:
                    third = third.pop()
                    self._check_number(quantity, third)

    @property
    def parse_files(self) -> "dict":
        return self._engine._parse_files

    def add_to_parser(self, path: "str", query: "str", data: "Any") -> None:
        self._engine.add_to_parser(path, query, data)

    @property
    def tarball(cls):
        return cls._engine.tarball

    # OVERRIDE
    def _load_from_backend(self, uids, expired=None) -> "Cuds":
        for uid in uids:
            if uid in self._registry:
                yield self._registry.get(uid)
            else:
                yield None

    # OVERRIDE
    def _apply_added(self, root_object, buffer) -> None:
        for obj in buffer.values():
            self._wrap(obj)

    # OVERRIDE
    def _apply_updated(self, root_object, buffer) -> None:
        pass

    # OVERRIDE
    def _apply_deleted(self, root_object, buffer) -> None:
        pass

    @property
    def config(self) -> dict:
        return self._engine._config

    @property
    def exit_code(self) -> int:
        return self._engine.exit_code



from ast import literal_eval
from distutils.util import strtobool
from typing import Any
from urllib.parse import unquote

from parsimonious.grammar import Grammar
from parsimonious.nodes import NodeVisitor

cgrammar = Grammar(
    r"""
    text            = ( function / dict / namedlist / namedlist5 / include / include2 / pair / comment / emptyline )*

    pair            = key space+ numerical endline inlinecom*
    function        = key space+ key space+ numerical endline inlinecom*
    include         = key space+ key space+ includestr endline inlinecom*
    include2        = key space+ includestr inlinecom*
    dict            = key emptyline+ space* lbr dictmember* rbr inlinecom*
    namedlist       = key emptyline+ space* lpar listmember* rpar endline inlinecom*
    namedlist2      = key space+ lpar tensormember2* rpar
    namedlist3      = key space+ key space+ lpar namedlistmember* rpar
    namedlist4      = key space+ key space+ lpar tensormember2* rpar endline
    namedlist5      = number emptyline+ space* lpar listmember* rpar inlinecom*

    listvalue       = namedlist3 / namedlist2 / tensor / dict
    dictvalue       = namedlist / namedlist4 / function / pair / dict

    list            = lsqbr tensormember+ rsqbr
    tensor          = lpar tensormember2+ rpar space*
    tensormember    = space* num+ space*
    tensormember2   = space* numerical+ space*
    dictmember      = space* dictvalue+ space*
    listmember      = space* listvalue* space*
    namedlistmember = space* numerical* space*
    numerical       = tensor / list / expression / num
    num             = number / bool / string

    key             = ~"[#\w]+"
    space           = ~"\s"
    endline         = ";"
    emptyline       = "\n"
    comment         = singlecom / multicom / inlinecom
    multicom        = ~"/\*[\w\W]+\*/"
    singlecom       = ~"//[\w\W][^\n]+//"
    inlinecom       = ~"//[\w\W][^\n]+"

    number          = ~"[-+]?[\d]*\.?[\d]+(?:[eE][-+]?[\d]+)?"
    string          = ~"[\w\"\$\#\/.\d\-\#]+"
    includestr      = ~"\"[\w\W][^\n;]+\""
    expression      = number tensor
    bool            = true / false

    true            = "true"
    false           = "false"

    lbr             = "{"
    rbr             = "}"
    lsqbr           = "["
    rsqbr           = "]"
    lpar            = "("
    rpar            = ")"
    """
)


dictgrammar = Grammar(
    r"""
    text        = textmember+
    textmember  = dict / dim / include / includepair / funcpair / tensorpair / pair / namedlist / namedlist4
    funcpair    = key colon space function
    includepair = key colon space lsqbr key comma* space* includestr rsqbr comma* space*
    tensorpair  = key colon space tensor
    pair        = key colon space pairmember+ comma* space*
    include     = includekey colon space includestr+ comma* space*
    dict        = key colon space lbr dictmember* rbr comma* space*
    namedlist   = key colon space lsqbr listmember* rsqbr comma* space*
    namedlist2  = key2 colon space lsqbr pairmember+ rsqbr comma* space*
    namedlist3  = key2 colon space lsqbr namedlistmember+ rsqbr comma* space*
    namedlist4  = number colon space lsqbr listmember+ rsqbr comma* space*
    dim         = dimkey colon space lsqbr dimmember+ rsqbr comma* space*

    function    = lsqbr key comma space pairmember+ rsqbr comma* space*
    listdict    = lbr pair rbr comma* space*
    listdict2   = lbr namedlist2 rbr comma* space*
    listdict3   = lbr dict rbr comma* space*
    tensor      = lsqbr tenmember+ rsqbr comma* space*

    dimmember   = number comma* space*
    tenmember   = pairmember comma* space*
    namedlistmember = value comma* space*
    pairmember  = tensor / value
    dictmember  = funcpair / namedlist / namedlist2 / namedlist3 / tensorpair / pair  / dict / dim
    listmember  = tensor / function / listdict / listdict2 / listdict3

    value       = bool  / expression / number / string / letter

    space       = ~"\s"
    key         = ~"[#\w][^\:\,\s]+"
    key2        = key* space* key*


    number      = ~"[-+]?[\d]*\.?[\d]+(?:[eE][-+]?[\d]+)?"
    string      = ~"[\w\"\$][^\}\,\]\[]+"
    expression  = ~"[\d\w\"\(\)][^\}\{\,\]\[]+"
    includestr  = ~"\"[\w\W][^\[\]]+\""
    letter      = ~"[A-Za-z]"
    bool        = true / false

    true        = "True"
    false       = "False"
    dimkey      = "dimensions"
    includekey  = "#include"

    lbr         = "{"
    rbr         = "}"
    lsqbr       = "["
    rsqbr       = "]"
    colon       = ":"
    comma       = ","
    quote       = "'"
    """
)


class CVisitor(NodeVisitor):
    def visit_text(self, node, visited_children):
        """Returns the overall output."""
        output = {}
        for child in visited_children:
            res = child.pop()
            if type(res) == dict:
                output.update(**res)
        return output

    def visit_pair(self, node, visited_children):
        """Gets each key/value pair, returns a tuple."""
        key, _, value, *_ = visited_children
        return {key: value}

    def visit_function(self, node, visited_children):
        """Gets each key/value pair, returns a tuple."""
        key1, _, key2, _, value, *_ = visited_children
        return {key1: [key2, value]}

    def visit_include(self, node, visited_children):
        return self.visit_function(node, visited_children)

    def visit_include2(self, node, visited_children):
        return self.visit_pair(node, visited_children)

    def visit_dict(self, node, visited_children):
        key, _, _, _, dictmember, *_ = visited_children
        return {
            key: {
                dictkey: value
                for member in dictmember
                for dictkey, value in member.items()
            }
        }

    def visit_namedlist(self, node, visited_children):
        key, _, _, _, listmember, *_ = visited_children
        output = []
        for member in listmember:
            output += member
        return {key: output}

    def visit_namedlist2(self, node, visited_children):
        key, _, _, listmember, _ = visited_children
        return {key: listmember}

    def visit_namedlist3(self, node, visited_children):
        key1, _, key2, _, _, listmember, *_ = visited_children
        if type(listmember) == list:
            if len(listmember) == 1:
                listmember = listmember.pop()
        return {f"{key1} {key2}": listmember}

    def visit_namedlist4(self, node, visited_children):
        return self.visit_namedlist3(node, visited_children)

    def visit_namedlist5(self, node, visited_children):
        key, _, _, _, listmember, *_ = visited_children
        output = []
        for member in listmember:
            output += member
        return {str(key): output}

    def visit_list(self, node, visited_children):
        _, member, _ = visited_children
        return member

    def visit_tensor(self, node, visited_children):
        _, member, *_ = visited_children
        if isinstance(member, list):
            if len(member) == 1 and (
                (isinstance(member[0], list) and node.expr_name == "tensor")
                or node.expr_name != "tensor"
            ):
                member = member.pop()
        return member

    def visit_dictmember(self, node, visited_children):
        return self.visit_tensor(node, visited_children)

    def visit_listmember(self, node, visited_children):
        return self.visit_list(node, visited_children)

    def visit_namedlistmember(self, node, visited_children):
        return self.visit_list(node, visited_children)

    def visit_tensormember(self, node, visited_children):
        return self.visit_tensor(node, visited_children)

    def visit_tensormember2(self, node, visited_children):
        return self.visit_tensormember(node, visited_children)

    def visit_numerical(self, node, visited_children):
        if visited_children:
            val = visited_children.pop()
        else:
            val = None
        return val

    def visit_num(self, node, visited_children):
        return self.visit_numerical(node, visited_children)

    def visit_dictvalue(self, node, visited_children):
        return self.visit_numerical(node, visited_children)

    def visit_listvalue(self, node, visited_children):
        return self.visit_numerical(node, visited_children)

    def visit_key(self, node, visited_children):
        return self.visit_string(node, visited_children)

    def visit_string(self, node, visited_children):
        return node.text

    def visit_number(self, node, visited_children):
        return literal_eval(node.text)

    def visit_bool(self, node, visited_children):
        return bool(strtobool(node.text))

    def generic_visit(self, node, visited_children):
        """The generic visit method."""
        return visited_children or node.text

    def visit_comment(self, node, visited_children):
        pass

    def visit_emptyline(self, node, visited_children):
        pass

    def visit_expression(self, node, visited_children):
        number, tensor = visited_children
        return f"{number}({' '.join(tensor)})"


class DictVisitor(NodeVisitor):
    def visit_text(self, node, visited_children):
        """Returns the overall output."""
        return self._gather_members(visited_children, endswith="\n")

    def visit_textmember(self, node, visited_children):
        return visited_children.pop()

    def visit_funcpair(self, node, visited_children):
        return self.visit_pair(node, visited_children)

    def visit_tensorpair(self, node, visited_children):
        return self.visit_pair(node, visited_children)

    def visit_pair(self, node, visited_children):
        key, _, _, member, *_ = visited_children
        return f"{key}\t{self._gather_members(member)};\n"

    def visit_dict(self, node, visited_children):
        key, _, _, _, member, *_ = visited_children
        return f"{key}\n{{\n{self._gather_members(member)}\n}}\n"

    def visit_includepair(self, node, visited_children):
        key1, _, _, _, key2, _, _, string, *_ = visited_children
        return f"{key1} {key2} {string};"

    def visit_listdict(self, node, visited_children):
        _, member, *_ = visited_children
        return member.replace(";", "")

    def visit_listdict2(self, node, visited_children):
        return self.visit_listdict(node, visited_children)

    def visit_listdict3(self, node, visited_children):
        _, member, *_ = visited_children
        return member

    def visit_namedlist(self, node, visited_children):
        key, _, _, _, member, *_ = visited_children
        return f"{key}\n(\n{self._gather_members(member)}\n);\n"

    def visit_namedlist2(self, node, visited_children):
        return self.visit_namedlist(node, visited_children)

    def visit_namedlist3(self, node, visited_children):
        return self.visit_namedlist(node, visited_children)

    def visit_namedlist4(self, node, visited_children):
        key, _, _, _, member, *_ = visited_children
        return f"{key}\n(\n{self._gather_members(member)}\n)\n"

    def visit_dim(self, node, visited_children):
        key, _, _, _, member, *_ = visited_children
        return f"{key}\t[{self._gather_members(member, endswith=' ')}];\n"

    def visit_tensor(self, node, visited_children):
        _, member, *_ = visited_children
        return f"({self._gather_members(member)})"

    def visit_function(self, node, visited_chilrdren):
        _, key, _, _, member, *_ = visited_chilrdren
        return f"{key}\t{self._gather_members(member)}"

    def visit_dictmember(self, node, visited_children):
        member = f"\n{visited_children.pop()}"
        return member.replace("\n", "\n\t")

    def visit_pairmember(self, node, visited_children):
        return visited_children.pop()

    def visit_listmember(self, node, visited_children):
        member = f"\n{visited_children.pop()}"
        return member.replace("\n", "\n\t")

    def visit_dimmember(self, node, visited_children):
        member, *_ = visited_children
        return member

    def visit_namedlistmember(self, node, visited_children):
        return self.visit_tenmember(node, visited_children)

    def visit_tenmember(self, node, visited_children):
        return f"\t{self.visit_dimmember(node, visited_children)}"

    def visit_value(self, node, visited_children):
        return f"\t{visited_children.pop()}"

    def visit_bool(self, node, visited_children):
        return visited_children.pop().lower()

    def visit_key2(self, node, visited_children):
        key1, _, key2 = visited_children
        if type(key1) == list:
            key1 = key1.pop()
        if type(key2) == list:
            key2 = key2.pop()
        return f"{key1} {key2}"

    def visit_include(self, node, visited_children):
        key, _, _, member, *_ = visited_children
        return f"{key}\t{self._gather_members(member)}\n"

    def visit_key(self, node, visited_children):
        return node.text

    def generic_visit(self, node, visited_children):
        """The generic visit method."""
        return visited_children or node.text

    def _gather_members(self, members: list, endswith: str = None):
        content = str()
        for child in members:
            content += str(child)
            if endswith:
                content += endswith
        return content


def check_type(key: str):
    if key.isdigit() or key.isdecimal():
        return literal_eval(key)
    else:
        try:
            return bool(strtobool(key))
        except Exception:
            return key


def serialize(dictionary: dict):
    content = str(dictionary)
    content = content[1:-1].replace("'", str())
    res = dictgrammar.parse(content)
    n = DictVisitor()
    return n.visit(res)


def run_parser(filepath):
    with open(filepath, "r+") as file:
        content = file.read()
    res = cgrammar.parse(content)
    n = CVisitor()
    return n.visit(res)


def get_sub(parse_dict, key, next_key):
    if type(parse_dict) == list:
        if type(key) != int:
            raise TypeError(f"Index {key} for list {parse_dict} must be of type <int>.")
        try:
            return_list = parse_dict[key]
        except IndexError:
            parse_dict += [0] * (key + 1 - len(parse_dict))
        if type(next_key) != str:
            parse_dict[key] = []
        else:
            parse_dict[key] = {}
        return parse_dict[key]
    elif type(parse_dict) == dict:
        if key not in parse_dict.keys():
            if type(next_key) != str:
                parse_dict[key] = []
            else:
                parse_dict[key] = {}
        return parse_dict[key]


def insert(parse_dict, key, value):
    if type(parse_dict) == list:
        if type(key) != int:
            raise TypeError(f"Index {key} for list {parse_dict} must be of type <int>.")
        try:
            parse_dict[key] = value
        except IndexError:
            parse_dict += [0] * (key + 1 - len(parse_dict))
            parse_dict[key] = value
    elif type(parse_dict) == dict:
        parse_dict[key] = value


def replace(filepath: str, path: str, value: Any) -> str:
    content = run_parser(filepath)
    replaced = deep_replace(content, path, value)
    return serialize(replaced)


def deep_replace(content: dict, path: str, value: Any) -> dict:
    keys = [unquote(key) for key in path.split(".")]
    parse_dict = content
    for ikey in range(len(keys)):
        key = keys[ikey]
        key = check_type(key)
        if ikey == len(keys) - 1:
            insert(parse_dict, key, value)
        else:
            next_key = keys[ikey + 1]
            next_key = check_type(next_key)
            parse_dict = get_sub(parse_dict, key, next_key)
    return content



import os
from pathlib import Path
from typing import Any, Dict

from pydantic import BaseSettings, Field

import osp.dictionaries.catalyticFoam as case

CASES = os.path.dirname(case.__file__)
STANDARD_CASE = os.path.join(CASES, "laminar_2D_ML")
DEFAULT_PKL = os.path.join(STANDARD_CASE, "ml_ExtraTrees_forCFD.pkl")


class ModelSettings(BaseSettings):
    catalyticfoam_standard_pkl: Path = Field(
        DEFAULT_PKL,
        description="Path to standard PKL-file with the trained surrogate model.",
    )

    catalyticfoam_standard_case: Path = Field(
        STANDARD_CASE, description="Path to standard case."
    )

    download_timeout: int = Field(
        10,
        description="Timeout when a file is downloaded by the pydantic model from a static url.",
    )


class CatalyticFoamSettings(BaseSettings):
    catalyticfoam_excluded_check_files: str = Field(
        "0/Rh_s_",
        description="Files to be excluded during the consistency-check of boundaries",
    )
    catalyticfoam_bash: Path = Field(
        "/opt/openfoam8/etc/bashrc",
        description="Bashrc to be sourced while launching the catalyticFOAM-solver.",
    )
    catalyticfoam_default_files: str = Field(
        "0/Ydefault | 0/thetaDefault",
        description="""Files holding default values for boundaryFields of chemical species,
        which were not defined by the user.""",
    )
    catalyticfoam_default_bounds: Dict[Any, Any] = Field(
        {"wedge": {"type": "wedge"}},
        description="""Python-dict setting the default data for bounaries in the `default_files`.
        The primary keys are string-matching with the patches available in the `constant/polyMesh/boundary`-file.""",
    )

    catalyticfoam_label_iri: str = Field(
        "http://emmo.info/emmo#EMMO_5d181bad_b072_4b33_82a8_774c20ca577d",
        description="""IRI for the annotation property to detect the OpenFOAM-syntax of a property.
        E.g. `fixedValue` for a Dirichlet boundary condition.""",
    )


class ReaxProSettings(ModelSettings, CatalyticFoamSettings):
    """General Reaxpro wrapper settings"""

    class Config:
        """Config-class for the general wrapper settings."""

        env_prefix = "REAXPRO_"



from typing import TYPE_CHECKING
from urllib.parse import quote, urlencode

from osp.core.namespaces import emmo
from osp.models.catalytic.utils import make_arcp

if TYPE_CHECKING:
    from typing import List, Union

    from osp.core.cuds import Cuds
    from osp.core.ontology import OntologyClass


def _make_internal_field(
    filepath: "str", value: "Union[int, float, List[float]]", oclass: "OntologyClass"
) -> "Cuds":
    quantity = oclass()
    # value
    path = ["internalField", "1"]
    if isinstance(value, float) or isinstance(value, int):
        value = _make_value(filepath, value, path)
    elif isinstance(value, list):
        value = _make_vector(filepath, value, path)
    else:
        raise TypeError("`value` must be type `int`, `float` or `list`.")
    quantity.add(value, rel=emmo.hasQuantityValue)
    # distribution
    iri = make_arcp(
        filepath,
        query=dict(
            jsonpath=[["internalField", "0"]],
        ),
    )
    distribution = emmo.UniformDistribution(iri=iri)
    quantity.add(distribution, rel=emmo.hasSpatialDirectPart)
    return quantity


def _make_boundary_field(
    filepath: "str",
    patch: "str",
    oclass_boundary: "OntologyClass",
    quantity: "OntologyClass" = None,
    value: "Union[int, float, List[float]]" = None,
) -> "Cuds":
    # type
    iri = make_arcp(filepath, query=dict(jsonpath=[["boundaryField", patch, "type"]]))
    bc = oclass_boundary(iri=iri)
    if value and quantity:
        quantity = quantity()
        # key
        if "gradient" in oclass_boundary.name.lower():
            key = "gradient"
        else:
            key = "value"
        # value
        path = ["boundaryField", patch, key]
        if isinstance(value, float) or isinstance(value, int):
            value = _make_value(filepath, value, path + ["1"])
        elif isinstance(value, list):
            value = _make_vector(filepath, value, path + ["1"])
        else:
            raise TypeError("`value` must be type `int`, `float` or `list`.")
        quantity.add(value, rel=emmo.hasQuantityValue)
        # distribution
        iri = make_arcp(
            filepath,
            query=dict(
                jsonpath=[
                    path + ["0"],
                ]
            ),
        )
        distribution = emmo.UniformDistribution(iri=iri)
        quantity.add(distribution, rel=emmo.hasSpatialDirectPart)
        bc.add(quantity, rel=emmo.hasSpatialDirectPart)
    return bc


def _make_value(
    filepath: "str",
    value: "Union[int, float]",
    path: "List[str]",
    oclass: "OntologyClass" = emmo.Real,
) -> "Cuds":
    iri = make_arcp(
        filepath,
        query=dict(jsonpath=[path]),
    )
    return oclass(iri=iri, hasNumericalData=value)


def _make_vector(filepath: "str", value: "List[float]", base: "List[str]") -> "Cuds":
    vector = emmo.Shape3Vector()
    components = []
    for i, val in enumerate(value):
        iri = make_arcp(
            filepath,
            query=dict(
                jsonpath=[
                    base + [str(i)],
                ]
            ),
        )
        components.append(emmo.Real(iri=iri, hasNumericalData=val))
    vector.add(components[0], rel=emmo.hasSpatialFirst)
    vector.add(components[2], rel=emmo.hasSpatialLast)
    components[0].add(components[1], rel=emmo.hasSpatialNext)
    components[1].add(components[2], rel=emmo.hasSpatialNext)
    return vector



"""Helper functions for OSP-utilities."""
import io
import json
import os
import tempfile
import warnings
from typing import Any, Dict, List, TextIO, Union
from uuid import uuid4

import requests
from minio import Minio
from minio.error import MinioException
from urllib3.response import HTTPResponse

import osp.dictionaries as cases
from osp.models.settings.general import ModelSettings

CASES = os.path.dirname(cases.__file__)


def get_minio() -> Minio:
    """Instantiate MinIO-client without `Depends`."""
    settings = ModelSettings()
    return Minio(
        settings.minio_endpoint,
        access_key=settings.minio_user.get_secret_value(),
        secret_key=settings.minio_password.get_secret_value(),
        secure=False,
    )


def _get_upload(filepath: str, uuid: str, minio_client: Minio) -> str:
    """Helper function for `depends_upload` and `get_upload`."""
    suffix = os.path.splitext(filepath)[-1]
    # Generate a unique UUID as the object key
    object_key = uuid or str(uuid4())

    # Upload the file to MinIO
    try:
        if not minio_client.bucket_exists(object_key):
            minio_client.make_bucket(object_key)
        minio_client.fput_object(
            object_key,
            object_key,
            filepath,
            metadata={"suffix": suffix},
        )
    except Exception as err:
        raise MinioException(err.args) from err
    return object_key


def _get_download(uuid: str, minio_client: Minio) -> HTTPResponse:
    """Helper function for `depends_download` and `get_download`."""
    file_data = io.BytesIO()
    if not minio_client.bucket_exists(uuid):
        raise MinioException("Bucket does not exist.")
    try:
        response = minio_client.get_object(
            uuid,
            uuid,
            file_data,
        )
    except Exception as err:
        raise MinioException(err) from err
    return response


def get_upload(file: Union[str, TextIO], uuid: str = None) -> str:
    """Upload file with minio client and return upload-id without `Depends`."""
    if hasattr(file, "filename"):
        file = file.filename
    elif hasattr(file, "name"):
        file = file.name
    minio_client = get_minio()
    return _get_upload(file, uuid, minio_client)


def get_download(uuid: str, as_file=False) -> Union[HTTPResponse, str]:
    """Return `io.BytesIO` from uuid through minio client without `Depends`."""
    minio_client = get_minio()
    response = _get_download(uuid, minio_client)
    if as_file:
        suffix = response.headers.get("x-amz-meta-suffix")
        with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as temp:
            temp.write(response.data)
            response = temp.name
    return response


def _get_example_json(example_json: str, standard_data: List) -> Dict[str, Any]:
    try:
        for uuid, filename in standard_data:
            try:
                get_download(uuid)
            except Exception:
                filepath = os.path.join(CASES, "example_data", filename)
                with open(filepath, "rb") as file:
                    get_upload(file, uuid=uuid)
    except Exception as error:
        warnings.warn(f"Error while uploading the example file to MinIO: {error.args}")
    return _get_example("ams", example_json)


def _get_example(folder: str, file: str) -> Dict[str, Any]:
    filepath = os.path.join(CASES, folder, file)
    with open(filepath, "r+") as handler:
        return json.load(handler)


def _download_file(url) -> str:
    settings = ModelSettings()
    response = requests.get(url, timeout=settings.download_timeout)
    return response.text
